  //@@@@ //@@@@@//@@ //@@//@@@@  //@@@@//@@@@@ //@@@@@@ -----------------------
 //@@//@@/@@ //@@/@@ //@@/@@//@@//@@//@@//@@/@@//@/@@/@  a companion of CSW2CDT
//@@    //@@    //@@ //@@   //@@/@@     //@@//@@ //@@    turning WAV audio into
//@@     //@@@@@//@@/@/@@//@@@@//@@     //@@//@@ //@@    CSW v1 files, coded by
//@@         //@@/@@@@@@@/@@   //@@     //@@//@@ //@@    Cesar Nicolas-Gonzalez
 //@@//@@/@@ //@@/@@@/@@@/@@//@@//@@//@@//@@/@@  //@@    since 2020/05/31-09:50
  //@@@@ //@@@@@//@@ //@@/@@@@@@ //@@@@//@@@@@  //@@@@ ------------------------

#define MY_VERSION "20240727"
#define MY_LICENSE "Copyright (C) 2020-2024 Cesar Nicolas-Gonzalez"

#define GPL_3_INFO \
	"This program comes with ABSOLUTELY NO WARRANTY; for more details" "\n" \
	"please see the GNU General Public License. This is free software" "\n" \
	"and you are welcome to redistribute it under certain conditions."

/* This notice applies to the source code of CSW2CDT and its binaries.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

Contact information: <mailto:cngsoft@gmail.com> */

#include <stdio.h> // fopen,printf...
#include <stdlib.h> // atoll,malloc...
#include <string.h> // memcmp,strcmp...

#define length(x) (sizeof(x)/sizeof*(x))
unsigned char buffer[512]; FILE *fi,*fo;
char *autosuffix(char *t,char *s,char *x) // return a valid path, with a new suffix if required
{
	if (t) return t; else if (!s) return NULL;
	if ((char*)buffer!=s) strcpy(buffer,s);
	if ((t=strrchr(buffer,'.'))&&(!(s=strrchr(buffer,
	#ifdef _WIN32
	'\\'
	#else
	'/'
	#endif
	))||t>s)) return strcpy(t,x),buffer; // override old suffix
	return strcat(buffer,x); // append new suffix
}

// I/O file operations, buffering and Intel lil-endian integer logic -------- //

#define fread1(t,i) fread((t),1,(i),fi)
#define fskip1(i) fseek(fi,(i),SEEK_CUR)
#define fwrite1(t,i) fwrite((t),1,(i),fo)
#define fput1(n) fputc((n),fo)
int fput4(int i) { fput1(i); fput1(i>>8); fput1(i>>16); return fput1(i>>24); } // non-buffered!

unsigned char tsrc[1<<12],ttgt[1<<12]; int isrc=0,ilen=0,itgt=0; // I/O file buffers
#define flush1() fwrite1(ttgt,itgt)
int frecv1(void) { while (isrc>=ilen) { if (isrc-=ilen,!(ilen=fread1(tsrc,sizeof(tsrc)))) return -1; } return tsrc[isrc++]; }
int frecv2(void) { int i=frecv1(); return i|(frecv1()<<8); }
int frecv4(void) { int i=frecv1(); i|=frecv1()<<8; i|=frecv1()<<16; return i|(frecv1()<<24); }
int fsend1(int i) { ttgt[itgt++]=i; if (itgt>=length(ttgt)) { if (!flush1()) return -1; else itgt=0; } return i; }
int fsend4(int i) { fsend1(i); fsend1(i>>8); fsend1(i>>16); return fsend1(i>>24); }

// the INFLATE procedure, simplified as much as possible; see also CPCEC-RT.H //

int puff_word,puff_bits; // stream cursor
int puff_len_c[16+288],puff_off_c[16+32]; // Huffman tables: 16 bases + N values
char puff_cnt_c[320]; // bit counts: 0..287 length codes + 288..319 offset codes

const int puff_len_k[2][32]={ // length constants; 0, 30 and 31 are reserved
	{ 0,3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83,99,115,131,163,195,227,258,0,0 },
	{ 0,0,0,0,0,0,0,0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4,  4,  5,  5,  5,  5,  0,0,0 }};
const int puff_off_k[2][32]={ // offset constants; 30 and 31 are reserved
	{ 1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073,4097,6145,8193,12289,16385,24577,0,0 },
	{ 0,0,0,0,1,1,2, 2, 3, 3, 4, 4, 5, 5,  6,  6,  7,  7,  8,  8,   9,   9,  10,  10,  11,  11,  12,   12,   13,   13,0,0 }};
const char puff_cnt_k[19]={ 16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15 }; // custom Huffman indices
void puff_default(void) // generate the default static Huffman bit counts
{
	int i=0; // lengths: 8 x144, 9 x112, 7 x24, 8 x8; offsets: 5 x32
	for (;i<144;++i) puff_cnt_c[i]=8; // 000..143 : 00110000...10111111.
	for (;i<256;++i) puff_cnt_c[i]=9; // 144..255 : 110010000..111111111
	for (;i<280;++i) puff_cnt_c[i]=7; // 256..279 : 0000000....0010111..
	for (;i<288;++i) puff_cnt_c[i]=8; // 280..287 : 11000000...11000111.
	for (;i<320;++i) puff_cnt_c[i]=5; // notice that the reserved codes take part in the calculations!
}
// Huffman-bitwise operations
int puff_recv(int n) // receive `n` bits from source; <0 ERROR, >=0 OUTPUT
{
	int w=puff_word,b; while (puff_bits<n) // we'll need `w` later
	{
		if ((b=frecv1())<0) return -1; // source underrun!
		else w+=b<<puff_bits,puff_bits+=8; // read byte from source
	}
	return puff_word=w>>n,puff_bits-=n,w&((1<<n)-1); // flush bits and return
}
int puff_decode(int *h) // receive code from Huffman source; <0 ERROR, >=0 OUTPUT
{
	for (int w=puff_word,b=puff_bits,l=1,i=0,base=0,code=0;;) // extract bits till they fit within an interval
	{
		for (int o;b;--b) // Huffman bits are stored backwards: "w" right-shifts, "code" left-shifts
			if (code+=w&1,w>>=1,code<base+(o=h[l])) // does the code fit the interval?
				return puff_word=w,puff_bits=(puff_bits-l)&7,h[16+i+(code-base)]; // yes!
			else i+=o,base=(base+o)<<1,code<<=1,++l; // calculate next interval
		if (!(b=16-l)||(w=frecv1())<0) return -1; // too many bits! source underrun!
		if (b>8) b=8; // read byte from source
	}
}
int puff_tables(int *h,char *l,int n) // generate Huffman input tables from canonical length table; 0 OK, !0 ERROR
{
	int o[16],i,a; for (i=0;i<16;++i) h[i]=0; // reset intervals
	for (i=0;i<n;++i) ++h[l[i]]; // calculate intervals
	if (h[0]==n) return 0; // empty table, nothing to do!
	for (a=i=1;i<16;++i) if ((a=(a<<1)-h[i])<0) return a; // <0 bad value!
	for (o[i=1]=0;i<15;++i) o[i+1]=o[i]+h[i]; // calculate base for each bit count
	for (i=0;i<n;++i) if (l[i]) h[16+o[l[i]]++]=i; // assign values to all symbols
	return a; // 0 if complete, >0 requires checking whether valh[0]+1==n, i.e. all lengths but one are zero
}
unsigned int puff_adler32(const unsigned char *t,int o) // calculates Adler-32 checksum of a block `t` of size `o`
	{ int j=1,k=0; while (o-->0) { if ((k+=(j+=*t++))>=65521) k%=65521,j%=65521; } return (k<<16)+j; }
unsigned char *puff_zlib(int *o) // unpacks a RFC1950 standard ZLIB source; returns a buffer (`o` receives its size), or NULL if something went wrong
{
	int z=1<<20,i,j,q; unsigned char *u=malloc(z*2),*t=u; if (!u) return NULL; // no memory!
	if (frecv1()!=120||(i=frecv1())&32||((30720%31)+i)%31) goto freenull; // bad ZLIB header!
	puff_word=puff_bits=0; do
	{
		if (q=puff_recv(1),!(i=puff_recv(2))) // stored block?
		{
			i=frecv2(),j=frecv2(); if (i<0||j<0) goto freenull; // source underrun!
			if (i+j!=0XFFFF) goto freenull; // bad value!
			for (;i;--i) *t++=j=frecv1(); if (j<0) goto freenull; // source underrun!
			// length zero is allowed: `...This completes the current deflate block and follows it with an empty stored block that is
			// three bits plus filler bits to the next byte, followed by four bytes (00 00 ff ff)...` http://www.zlib.net/manual.html
			puff_word=puff_bits=0; // drop any bits left
			if ((i=t-u)>=z) { if (u=realloc(u,(z*=2)*2)) t=u+i; else return NULL; } // no memory!
		}
		else if (i>0&&i<3) // packed block?
		{
			if (i==1) // default Huffman trees?
			{
				puff_default(); // default static Huffman tables
				puff_tables(puff_len_c,puff_cnt_c,288); // THERE MUST BE EXACTLY 288 LENGTH CODES!!
				puff_tables(puff_off_c,puff_cnt_c+288,32); // DITTO: THERE MUST BE 32 OFFSET CODES!
			}
			else // custom Huffman trees!
			{
				int len,off,a; if ((len=puff_recv(5)+257)>286||(off=puff_recv(5)+1)>30||(j=puff_recv(4)+4)<4)
					goto freenull; // bad match/range/huffman values!
				for (a=0,i=0;i<j;) a|=puff_cnt_c[puff_cnt_k[i++]]=puff_recv(3); // read bit counts for the Huffman-encoded header
				while (i<19) puff_cnt_c[puff_cnt_k[i++]]=0; // padding: clear the remainder of this header
				if (a<0||puff_tables(puff_len_c,puff_cnt_c,19)) goto freenull; // bad values! invalid table!
				for (len+=off,i=0;i<len;)
					if ((a=puff_decode(puff_len_c))<16) // literal?
						if (a<0) goto freenull; else puff_cnt_c[i++]=a; // bad value!
					else
					{
						if (a<17) // 16: repeat the last value 3..6 times
							if (i<1) goto freenull; else j=puff_cnt_c[i-1],a=3+puff_recv(2); // bad value!
						else // 17 or 18: store zero 3..10 or 11..138 times
							j=0,a=a==17?3+puff_recv(3):11+puff_recv(7);
						if (a<3||i+a>len) goto freenull; // source underrun! bad value!
						do puff_cnt_c[i++]=j; while (--a);
					}
				if ((len-=off,puff_tables(puff_len_c,puff_cnt_c,len)&&len-puff_len_c[0]!=1)|| // bad length table!
					(puff_tables(puff_off_c,puff_cnt_c+len,off)&&off-puff_off_c[0]!=1)) goto freenull; // bad offset table!
			}
			for (;;) // beware: custom trees cannot generate length codes >=286 or offset codes >=30, but default trees can!
			{
				if ((i=puff_decode(puff_len_c))<256) // literal?
					if (i<0) goto freenull; else *t++=i; // bad value!
				else if (i-=256) // length:offset pair?
				{
					if (i>=30||(i=puff_len_k[0][i]+puff_recv(puff_len_k[1][i]))<0|| // bad value! source underrun!
						(j=puff_decode(puff_off_c))<0||j>=30|| // bad value!
						(j=puff_off_k[0][j]+puff_recv(puff_off_k[1][j]))>t-u) goto freenull; // source underrun!
					unsigned char *s=t-j; do *t++=*s++; while (--i);
				}
				else break; // end of block!
				if ((i=t-u)>=z) { if (u=realloc(u,(z*=2)*2)) t=u+i; else return NULL; } // no memory!
			}
		}
		else goto freenull; // unsupported block!
	}
	while (!q);
	//FILE *ff=fopen("debug!","wb"); if (ff) fwrite(u,1,t-u,ff),fclose(ff);
	i=puff_adler32(u,t-u); j=frecv1()<<24; j+=frecv1()<<16; j+=frecv1()<<8;
	if ((j+=frecv1())!=i) { freenull: return free(u),NULL; } // *!* GOTO! bad checksum!
	return *o=t-u,u; // OK, output length
}

// the main procedure proper ------------------------------------------------ //

const char csw24b[]="Compressed Square Wave\032\001",ok_bytes[]="%d:%d bytes.\n",bad_target[]="error: cannot create target!\n";
int plusatoi(char *s) { return atoi(*s=='+'?s+1:s); } // handle "+1234" as "1234"
int readmmss(char *s) { int h=0,l=0,c; while ((c=*s++)>='0'&&c<='9') l=l*10+c-'0'; if (c==':') { h=l*60,l=0; while ((c=*s++)>='0'&&c<='9') l=l*10+c-'0'; } return c?-1:h+l; }

int main(int argc,char *argv[])
{
	int h,i,j,k,l; char *si=NULL,*so=NULL;
	int flag_m=0,flag_e=0,flag_lr=0,flag_t=0,flag_n=0,filter_l=0,filter_h=0,hz,mmss_init=0,mmss_exit=-1,flag_o=0;
	for (i=1;i<argc;++i)
		if (argv[i][0]=='-')
			if (argv[i][1])
				#define FETCH_INT(t,l,h) ((argv[i][++j]||(j=0,++i<argc))?t=plusatoi(&argv[i][j]),j=-1,t>=l&&t<h:0)
				for (j=0;++j>0&&i<argc&&argv[i][j];)
					switch (argv[i][j])
					{
						case 'f':
							if (!FETCH_INT(filter_l,0,96000))
								i=argc; // help!
							else if (++i>=argc||!FETCH_INT(filter_h,0,96000))
								i=argc; // help!
							if (filter_l&&filter_h&&filter_l>filter_h)
								k=filter_l,filter_l=filter_h,filter_h=k; // swap if L > H
							break;
						case 'm':
							if (!FETCH_INT(flag_m,-127,+127))
								i=argc; // help!
							break;
						case 'e':
							if (!FETCH_INT(flag_e,0,127))
								i=argc; // help!
							break;
						case 'l':
							flag_lr=-1;
							break;
						case 'r':
							flag_lr=+1;
							break;
						case 'n':
							flag_n=1;
							break;
						case 't':
							if (!FETCH_INT(flag_t,0,100))
								i=argc; // help!
							break;
						case 'o':
							if (!FETCH_INT(flag_o,8000,96000))
								i=argc; // help!
							break;
						case '-':
							if (i+1<argc&&(j=readmmss(&argv[i][j+1]))>=0&&(mmss_init=j,(j=readmmss(argv[++i]))>=0))
								mmss_exit=j,j=-1;
							else
								i=argc; // help!
							break;
						case 'h':
						default:
							i=argc; // help!
							break;
					}
			else
				i=argc; // help!
		else if (!si)
			si=argv[i];
		else if (!so)
			so=argv[i];
		else i=argc; // help!
	if (i>argc||!si)
		return printf(
			"CSW0 " MY_VERSION " " MY_LICENSE "\n"
			"\n"
			"encoding from WAV to CSW:\n"
			"\tcsw0 [option..] source.wav [target.csw]\n"
			"\t-f L H\tenable pass-band filter, in Hz\n"
			"\t-m N\tset middle point between -127 and 127\n"
			"\t-e N\tset error margin between 0 and 127\n"
			"\t-l\tleft channel only\n"
			"\t-r\tright channel only\n"
			"\t-n\tnegative polarity\n"
			"\t--M1:S1 M2:S2\tclip from M1 min S1 sec to M2 min M2 sec\n"
			"decoding from CSW to WAV:\n"
			"\tcsw0 [option..] source.csw [target.wav]\n"
			"\t-t N\tset signal shape between 0 and 100 (square-triangle)\n"
			"\t-n\tnegative polarity\n"
			"\n" GPL_3_INFO "\n"
		),1;
	if (!(fi=fopen(si,"rb")))
		return fprintf(stderr,"error: cannot open source!\n"),1;
	if (*buffer=0,fread1(buffer,12),!memcmp(buffer,"RIFF",4)&&!memcmp(&buffer[8],"WAVE",4)) // encoding?
	{
		while (j=frecv4(),k=frecv4(),k>=0&&j!=0x20746D66) // "fmt "
			isrc+=(k+1)&~1; // RIFF even-padding
		if (k<16)
			return fclose(fi),fprintf(stderr,"error: source lacks header!\n"),1;
		int stereo=(frecv4()>>17)&1; // channels-1
		hz=frecv4();
		frecv4(); // actual byte rate...
		int datasize,byteskip=((frecv4()>>19)&3)-1,bytesize=byteskip+1; // bitdepth 8, 16 or 24 (!) -> byteskip = 0,1,2 (unused bytes per sample); bytesize = 1,2,3 (sample size in bytes)
		isrc+=((k+1)&~1)-16; // RIFF even-padding
		while (j=frecv4(),datasize=frecv4(),datasize>=0&&j!=0x61746164) // "data"
			isrc+=datasize;
		if (datasize<0)
			return fclose(fi),fprintf(stderr,"error: source lacks body!\n"),1;
		if (!(fo=fopen(autosuffix(so,si,".csw"),"wb")))
			return fclose(fi),fprintf(stderr,bad_target),1;
		flag_m+=128; fwrite1(csw24b,24);
		fsend4((flag_o?flag_o:hz)*256+1+256*256*256); // forge sample rate on command
		if (stereo) if (flag_lr<0) isrc-=bytesize; // avoid right channel
		if (mmss_init>0) { mmss_init*=(stereo+1)*bytesize*hz; isrc+=mmss_init,datasize-=mmss_init; } // source start...
		if (mmss_exit>0) { mmss_exit*=(stereo+1)*bytesize*hz; if (datasize>mmss_exit) datasize=mmss_exit; } // ...and end
		#define GETSAMPLE() (stereo? \
			byteskip?(datasize-=2*bytesize,isrc+=byteskip,i=frecv1()^128,isrc+=byteskip,i=(((frecv1()^128)+i)>>1)): \
				(datasize-=2,i=((frecv1()+frecv1())>>1)): \
			byteskip?(datasize-=bytesize,isrc+=byteskip,i=frecv1()^128): \
				(--datasize,i=frecv1()))
		#define SENDCHUNK do{ if (j) { if (j>255) fsend1(0),fsend4(j); else fsend1(j); j=0; } }while(0)
		GETSAMPLE(); fsend4(flag_n^(k=i>flag_m)); // store 1st signal!
		j=0; l=flag_m-flag_e; h=flag_m+flag_e;
		const float TWOPI=6.28318530717958647692528676655901;
		if (hz*filter_l>0)
		{
			if (hz*filter_h>0) // band-pass filter: H and L nonzero
			{
				int g=0; float h1=0,h2=0,high_alpha=1.0/(TWOPI*filter_l/hz+1.0),
					loww_alpha=(TWOPI*filter_h/hz)/(TWOPI*filter_h/hz+1.0);
				do
				{
					i-=flag_m; h2=high_alpha*(h2+i-g); g=i; // high-pass 1st
					h1=h1+loww_alpha*(h2-h1); // loww-pass 2nd
					if (k?(h1<=-flag_e):(h1>flag_e))
						{ SENDCHUNK; k=!k; }
				}
				while (++j,GETSAMPLE(),datasize>0&&ilen>0);
			}
			else // high-pass filter: H zero, L nonzero
			{
				int g=0; float hh=0,high_alpha=1.0/(TWOPI*filter_l/hz+1.0);
				do
				{
					i-=flag_m; hh=high_alpha*(hh+i-g); g=i; // high-pass
					if (k?(hh<=-flag_e):(hh>flag_e))
						{ SENDCHUNK; k=!k; }
				}
				while (++j,GETSAMPLE(),datasize>0&&ilen>0);
			}
		}
		else if (hz*filter_h>0) // low-pass filter: H nonzero, L zero
		{
			float hh=0,loww_alpha=(TWOPI*filter_h/hz)/(TWOPI*filter_h/hz+1.0);
			do
			{
				i-=flag_m; hh=hh+loww_alpha*(i-hh); // loww-pass
				if (k?(hh<=-flag_e):(hh>flag_e))
					{ SENDCHUNK; k=!k; }
			}
			while (++j,GETSAMPLE(),datasize>0&&ilen>0);
		}
		else // no filter
		{
			do
			{
				if (k?(i<=l):(i>h))
					{ SENDCHUNK; k=!k; }
			}
			while (++j,GETSAMPLE(),datasize>0&&ilen>0);
		}
		SENDCHUNK; flush1(); // flush!
		printf(ok_bytes,(int)ftell(fi),(int)ftell(fo));
		return fclose(fi),fclose(fo),0;
	}
	char csw2=0; // CSW2 support is read-only!
	if (fread1(&buffer[12],(32-12))==(32-12)&&(!memcmp(buffer,csw24b,24)||(csw2=1,!memcmp(buffer,csw24b,23)))) // decoding?
	{
		if (csw2)
			{ hz=buffer[25]+buffer[26]*256+buffer[27]*65536; fread1(&buffer[32],20); csw2=buffer[33]==2; k=buffer[34]; fskip1(buffer[35]); }
		else
			{ hz=buffer[25]+buffer[26]*256+(buffer[27]-1)*65536; k=buffer[28]; }
		unsigned char *z0=NULL,*z1=NULL; int z9=0;
		if (csw2&&!(z0=z1=puff_zlib(&z9))) return fclose(fi),fprintf(stderr,"error: cannot unpack source!\n"),1;
		if (!(fo=fopen(autosuffix(so,si,".wav"),"wb"))) // this call destroys `buffer`!
			return fclose(fi),fprintf(stderr,bad_target),1;
		fwrite1("RIFFFFFFWAVEfmt \020\000\000\000\001\000\001\000~~~~~~~~\001\000\010\000dataaaaa",44);
		int g=2+(hz-1)/1000; // i.e. almost 2 ms; perhaps we should let the user set the limit
		for (k=((flag_n^k)&1)?32:224;(i=(z1?--z9>=0?*z1++:-1:frecv1()))>=0;)
		{
			k^=192; if (!i) i=z1?(z9-=4)>=0?z1+=4,(z1[-1]<<24)|(z1[-2]<<16)|(z1[-3]<<8)|z1[-4]:-1:frecv4();
			// long edges become flat at the end!
			l=0; if (i>g) l=i-g,i=g;
			// rising slope
			int g1=i*flag_t/200,g2=(i*flag_t+100)/200; i-=g1+g2;
			j=k-128; h=g1>>1; for (int n=1;n<=g1;++n) fsend1(128+(j*n+h)/g1);
			// ceiling, i.e. 128+96 / 128-96
			j=k; while (i>0) fsend1(j),--i;
			// falling slope
			j=k-128; h=g2>>1; for (int n=g2;n>=1;--n) fsend1(128+(j*n+h)/g2);
			// floor
			while (l) fsend1(128),--l;
		}
		if (z0) free(z0); // release the unpacked CSW2 source, if any
		{ if (itgt&1) fsend1(0); } flush1(); // RIFF even-padding + flush!
		printf(ok_bytes,(int)ftell(fi),i=(int)ftell(fo));
		fseek(fo, 4,SEEK_SET); fput4(i-8); // file size-8 = chunk size+36
		if (flag_o) hz=flag_o; // forge sample rate on command
		fseek(fo,24,SEEK_SET); fput4(hz); fput4(hz); // clock + bandwidth
		fseek(fo,40,SEEK_SET); fput4(i-8-36); // chunk size = filesize-44
		return fclose(fi),fclose(fo),0;
	}
	return fprintf(stderr,"error: unknown source type!\n"),1;
}
